<?php

namespace Drupal\swapcard_content\Form;

use Symfony\Component\DependencyInjection\ContainerInterface;

use Drupal\Core\Security\TrustedCallbackInterface;

use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\Queue\QueueWorkerManager;

use Drupal\Component\Utility\NestedArray;
use Drupal\Component\Utility\Html;

use Drupal\swapcard\Form\SwapcardConfigForm;
use Drupal\swapcard\Plugin\SwapcardPluginManager;

/**
 * SwapcardContentConfigForm object definition.
 */
class SwapcardContentConfigForm extends SwapcardConfigForm implements TrustedCallbackInterface {

  /**
   * Drupal\Core\Queue\QueueWorkerManager definition.
   *
   * @var \Drupal\Core\Queue\QueueWorkerManager
   */
  protected $queueManager;

  /**
   * Constructs \Drupal\swapcard_content\Form\SwapcardContentConfigForm object.
   *
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The factory for configuration objects.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   Entity type manager interface.
   * @param \Drupal\Core\Cache\CacheBackendInterface $cache
   *   Cache interface.
   * @param \Drupal\swapcard\Plugin\SwapcardPluginManager $swapcard_plugin_manager
   *   Swapcard plugin manager.
   * @param \Drupal\Core\Queue\QueueWorkerManager $queue_manager
   *   QueueWorkerManager instance.
   */
  public function __construct(ConfigFactoryInterface $config_factory, EntityTypeManagerInterface $entity_type_manager, CacheBackendInterface $cache, SwapcardPluginManager $swapcard_plugin_manager, QueueWorkerManager $queue_manager) {
    parent::__construct($config_factory, $entity_type_manager, $cache, $swapcard_plugin_manager);
    $this->queueManager = $queue_manager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('config.factory'),
      $container->get('entity_type.manager'),
      $container->get('cache.default'),
      $container->get('plugin.manager.swapcard'),
      $container->get('plugin.manager.queue_worker')
    );
  }

  /**
   * {@inheritdoc}
   */
  public static function trustedCallbacks() {
    return [
      'processEntities',
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {

    $form = parent::buildForm($form, $form_state);

    $config = $this->config('swapcard_content.settings');

    $form['swapcard_content'] = [
      '#type' => 'details',
      '#title' => $this->t('Swapcard Content settings'),
      '#open' => TRUE,
      '#tree' => TRUE,
    ];

    // Container for multiple Events.
    $events_id = Html::getId('swapcard-events-wrapper');
    $form['swapcard_content']['events_container'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('Events'),
      '#description' => $this->t('Set IDs for Swapcard events you wish to include in the flow.'),
      '#tree' => TRUE,
    ];

    $form['swapcard_content']['events_container'] += $this->multipleEvents($form_state, $events_id, 'events', 'Event');

    $form['swapcard_content']['event_content'] = [
      '#type' => 'details',
      '#title' => $this->t('Event content'),
      '#description' => $this->t("Select event's content entities to import for each event."),
      '#open' => FALSE,
    ];

    $form['swapcard_content']['event_content']['entities'] = [
      '#type' => 'checkboxes',
      '#title' => $this->t('Event entities'),
      '#default_value' => $config->get('entities'),
      '#description' => $this->t("Note that, as per Swapcard design <em>Speakers</em> are child property of sessions."),
      '#pre_render' => [
        [
          get_class($this), 'processEntities',
        ],
      ],
      '#options' => [
        'swapcard_exhibitor' => $this->t('Exhibitors'),
        'swapcard_session' => $this->t('Sessions'),
        'swapcard_speaker' => $this->t('Speakers'),
      ],
    ];

    $node_types = $config->get('node_types');

    $default_fields = [];
    $custom_fields = [];

    foreach ($node_types as $bundle => $fields) {

      foreach ($fields as $drupal_field_name => $swapcard_field) {
        if (isset($swapcard_field['id'])) {
          if (isset($swapcard_field['custom'])) {
            $field_id = 'node.' . $bundle . '.' . $drupal_field_name;
            $custom_fields[] = [
              'drupal_field_name' => $field_id,
              'swapcard_field_id' => $swapcard_field['id'],
            ];
          }
          else {
            $default_fields[$drupal_field_name] = $swapcard_field['id'];
          }
        }
      }
    }

    // Container for multiple items.
    $form['swapcard_content']['event_content']['fields'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('Default fields'),
      '#description' => $this->t('Handle fields with pre-defined values in Swapcard app, a dropdown type mostly. With specifying its ID found in Swapcard app here, we are able to pull that pre-defined values and set them as "allowed_values" for these fields/dropdowns in Drupal. That way those fields are always synchronised with remote configuraton.'),
    ];

    $form['swapcard_content']['event_content']['fields']['field_swapcard_session_type'] = [
      '#type' => 'textfield',
      '#title' => $this->t("Session's <em>Type</em> field ID"),
      '#default_value' => $default_fields['field_swapcard_session_type'] ?? NULL,
    ];

    $form['swapcard_content']['event_content']['fields']['field_swapcard_location'] = [
      '#type' => 'textfield',
      '#title' => $this->t("Session's <em>Location</em> field ID"),
      '#default_value' => $default_fields['field_swapcard_location'] ?? NULL,
    ];

    $custom_fields_id = Html::getId('swapcard-fields-wrapper');
    $form['swapcard_content']['event_content']['custom_fields'] = $this->multipleFields($form_state, $custom_fields_id, 'fields', 'Custom field', $custom_fields);

    $default_uid = NULL;

    if ($config->get('default_uid')) {
      $default_uid = $config->get('default_uid');
    }
    else {
      $default_uid = $form_state->getValue(['swapcard_content', 'default_uid']) ? $form_state->getValue([
        'swapcard_content',
        'default_uid',
      ]) : NULL;
    }

    $node_author = $default_uid ? $this->entityTypeManager->getStorage('user')->load($default_uid) : NULL;

    $form['swapcard_content']['default_uid'] = [
      '#type' => 'entity_autocomplete',
      '#title' => $this->t('Swapcard nodes author'),
      '#description' => $this->t('Select user which becomes author of Swapcard nodes.'),
      '#target_type' => 'user',
      '#default_value' => $node_author,
      '#required' => TRUE,
    ];

    $form['swapcard_content']['cron'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Sync on cron'),
      '#default_value' => $config->get('cron'),
      '#description' => $this->t('Check this to run synchronisation on each cron.'),
    ];

    $form['swapcard_content']['advanced'] = [
      '#type' => 'details',
      '#title' => $this->t('Advanced'),
      '#description' => $this->t('Some advanced features, mostly for testing and debugging, but also for purging data.'),
      '#open' => FALSE,
    ];

    $form['swapcard_content']['advanced']['purge'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Purge entities'),
      '#default_value' => $config->get('purge'),
      '#description' => $this->t('When checked the obsolete entities, such as Exhibitors, Sessions as well as possibly Speakers, would be deleted. Each of such does not have a relation and/or existence on Swapcard anymore, but previously had, so we would like want to delete it from our image here in Drupal. Think of it as reverse check also may be resources demanding. <strong>Use AT YOUR OWN risk, it is in <em>alpha</em> stage by now</strong>, especially when it is about Speakers.'),
    ];

    // Kill switch, when true does not save anything in Drupal.
    // A big note - when using this QueueWorker still writes data in database
    // and will try to deliver all of it on the next cron.
    $form['swapcard_content']['advanced']['dry_run'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Dry run'),
      '#default_value' => $config->get('dry_run'),
      '#description' => $this->t('If true no content will be saved in Drupal yet all the other operations will run.'),
    ];

    $form['swapcard_content']['advanced']['dump'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Dump API response (YML)'),
      '#default_value' => $config->get('dump'),
      '#description' => $this->t('When checked "response" data which came from API will be saved into YML file in <em>public://swapcard/tests</em> folder. This is disabled when the next option is checked (loading tests).'),
      '#id' => 'swapcard-content-dump',
      '#states' => [
        'enabled' => [
          ':input[id="swapcard-content-test"]' => ['checked' => FALSE],
        ],
      ],
    ];

    $form['swapcard_content']['advanced']['test'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Use dumped response (YML)'),
      '#default_value' => $config->get('test'),
      '#description' => $this->t('When checked "response" data will come locally, from YML files previously dumped in <em>public://swapcard/tests</em>. This is disabled when the previous option is checked (dumping tests). This is the way to avoid too many calls to API but also it may come very benefitial when it is about deleting stuff, such as values for fields and/or even some of the main 3 entities from parent event.'),
      '#id' => 'swapcard-content-test',
      '#states' => [
        'disabled' => [
          ':input[id="swapcard-content-dump"]' => ['checked' => TRUE],
        ],
      ],
    ];

    // Synchronise data with remote. Run Queue in a batch.
    $form['actions']['batch_sync'] = [
      '#type' => 'submit',
      '#submit' => ['::batchSyncSubmit'],
      '#value' => $this->t('Sync'),
    ];

    // Purge any obsolete data (if event was removed previously for instance).
    $form['actions']['purge_data'] = [
      '#type' => 'submit',
      '#submit' => ['::purgeDataSubmit'],
      '#value' => $this->t('Purge'),
      '#disabled' => TRUE,
    ];

    // Clear any leftover queue tables.
    $form['actions']['clear'] = [
      '#type' => 'submit',
      '#submit' => ['::clearQueuesSubmit'],
      '#value' => $this->t('Clear'),
    ];

    // Buttons explanation.
    $buttons_markup = $this->t("<p><strong>Save configuration</strong> does default operation but in case when <em>Event fields</em> are set submit will also refresh those fields' values.</p><p><strong>Sync</strong> button will synchronise data in Drupal, mirroring what is currently on Swapcard app for Events defined above, as well as selected Entities for each. </p><p><strong>Purge</strong> is disabled and possibly meant for purging all data related to some event that was just removed, using <em>remove event</em> button above. Attention - this will delete all of the related Exhibitors and Sessions, as well as some Speakers (orphans) that do not have any parent Session anymore. Likewise orphan Media entities (images) will be removed.</p><p><strong>Clear</strong> button will clear any leftover queues in database tables, just these related to this module and starting with <em>swapcard_content_queue_</em>.</p>");

    $form['actions_description'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('Actions'),
      '#description' => $buttons_markup,
    ];

    return $form;
  }

  /**
   * Process entities checkboxes.
   *
   * Basically to disable speakers when no sessions selected.
   *
   * @param array $element
   *   Checkboxes form element.
   */
  public static function processEntities(array $element) {
    $element['swapcard_session']['#id'] = 'swapcard-content-session-id';
    $element['swapcard_speaker']['#states'] = [
      'enabled' => [
        ':input[id="swapcard-content-session-id"]' => ['checked' => TRUE],
      ],
    ];
    return $element;
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {

    $config = $this->configFactory->getEditable('swapcard_content.settings');

    // Take care of checkboxes and user.
    $default_uid = $form_state->getValue(['swapcard_content', 'default_uid']);
    $cron = $form_state->getValue(['swapcard_content', 'cron']);
    $dump = $form_state->getValue(['swapcard_content', 'advanced', 'dump']);
    $test = $form_state->getValue(['swapcard_content', 'advanced', 'test']);
    $dry_run = $form_state->getValue(['swapcard_content', 'advanced', 'dry_run']);
    $purge = $form_state->getValue(['swapcard_content', 'advanced', 'purge']);

    $config->set('default_uid', $default_uid);
    $config->set('cron', $cron);
    $config->set('dump', $dump);
    $config->set('test', $test);
    $config->set('dry_run', $dry_run);
    $config->set('purge', $purge);

    // Take care of event child entities checboxes.
    $event_content = $form_state->getValue([
      'swapcard_content',
      'event_content',
      'entities',
    ]);
    $save_entities = array_filter($event_content, function ($value) {
      return is_string($value) ?? $value;
    });

    $config->set('entities', $save_entities);

    // Take care of event fields entries.
    $default_fields = $form_state->getValue([
      'swapcard_content',
      'event_content',
      'fields',
    ]);

    $custom_fields = $form_state->getValue([
      'swapcard_content',
      'event_content',
      'custom_fields',
      'fields',
    ]);

    $filtered_fields = NULL;
    if (!empty($custom_fields)) {
      $filtered_fields = array_filter($custom_fields, function ($e) {
        return is_numeric($e);
      }, ARRAY_FILTER_USE_KEY);
    }

    $node_types = $config->get('node_types');
    $update_fields = $this->processFields($default_fields, $filtered_fields, $node_types);
    $config->set('node_types', $update_fields['update_fields']);

    // Take care of events entries.
    $events = $form_state->getValue([
      'swapcard_content',
      'events_container',
      'events',
    ]);
    $save_events = [];

    if (!empty($events)) {
      foreach ($events as $delta => $event_id) {
        if (is_numeric($delta) && !empty($event_id)) {
          $save_events[$delta] = trim($event_id);
        }
      }

      // REFRESH defined fields' allowed_values function.
      if (!empty($save_events) && !empty($update_fields['field_ids'])) {
        $fields_manager = $this->queueManager->createInstance('swapcard_content_queue_swapcard_fields');
        $fields_manager->queueFactory->get('swapcard_content_queue_swapcard_fields')->createItem($update_fields['field_ids']);
        $fields_manager->queueUi->batch(['swapcard_content_queue_swapcard_fields']);
      }
    }

    // Save config and return parent method.
    $config->set('events', $save_events);
    $config->save();
    return parent::submitForm($form, $form_state);
  }

  /**
   * Process bulk synchronisation with data from remote.
   *
   * @param array $form
   *   The form where the settings form is being included in.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   */
  public function batchSyncSubmit(array &$form, FormStateInterface $form_state) {
    $plugin_manager = $this->queueManager->createInstance('swapcard_content_queue_swapcard_event');
    $plugin_manager->queueFactory->get('swapcard_content_queue_swapcard_request')->createItem(['batch' => TRUE]);
    $plugin_manager->queueUi->batch(['swapcard_content_queue_swapcard_request']);
  }

  /**
   * Remove all queue tables, to start over.
   *
   * @param array $form
   *   The form where the settings form is being included in.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   */
  public function clearQueuesSubmit(array &$form, FormStateInterface $form_state) {
    $plugin = $this->queueManager->createInstance('swapcard_content_queue_swapcard_event');
    $queues = [
      'swapcard_content_queue_swapcard_event',
      'swapcard_content_queue_swapcard_exhibitor',
      'swapcard_content_queue_swapcard_session',
      'swapcard_content_queue_swapcard_speakers',
      'swapcard_content_queue_swapcard_fields',
      'swapcard_content_queue_swapcard_media',
      'swapcard_content_queue_swapcard_delete',
    ];

    foreach ($queues as $queue) {
      $plugin->queueFactory->get($queue)->deleteQueue();
    }
    $plugin->messenger->addStatus($this->t('Queue tables cleared.'));
  }

  /**
   * Purge-obsolete submit.
   *
   * @todo this one.
   *
   * @param array $form
   *   The form where the settings form is being included in.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   */
  public function purgeDataSubmit(array &$form, FormStateInterface $form_state) {
    $plugin = $this->queueManager->createInstance('swapcard_content_queue_swapcard_event');
    $plugin->messenger->addWarning($this->t('Still TODO!'));
  }

  /**
   * Callback for all ajax actions.
   *
   * Returns parent container element for each group.
   */
  public static function ajaxCallback(array &$form, FormStateInterface $form_state) {
    $trigger = $form_state->getTriggeringElement();
    $parents = array_slice($trigger['#parents'], 0, -1);
    $element = NestedArray::getValue($form, $parents);
    return $element;
  }

  /**
   * Add multiple element Submit callback.
   */
  public static function addItemSubmit(array &$form, FormStateInterface $form_state) {
    $trigger = $form_state->getTriggeringElement();
    $key = str_replace('op-', '', $trigger['#name']);
    $form_state->set('multiple_keys', $key);
    $num_items = $form_state->get('num_' . $key);
    $delta = $num_items + 1;
    $form_state->set('num_' . $key, $delta);
    $form_state->setRebuild(TRUE);
  }

  /**
   * Remove multiple element Submit callback.
   */
  public static function removeItemSubmit(array &$form, FormStateInterface $form_state) {
    $trigger = $form_state->getTriggeringElement();
    $key = str_replace('op-', '', $trigger['#name']);
    $form_state->set('multiple_keys', $key);
    $num_items = $form_state->get('num_' . $key);
    $delta = $num_items - 1;
    $form_state->set('num_' . $key, $delta);
    $form_state->setRebuild(TRUE);
  }

  /**
   * Process fields.
   *
   * @todo this one.
   *
   * @param array $default_fields
   *   An array of default fields.
   * @param array $filtered_fields
   *   An array of custom fields.
   * @param array $node_types
   *   Default config's node_types array.
   *
   * @return array
   *   An array containing fields to be updated.
   */
  private function processFields(array $default_fields, array $filtered_fields, array $node_types) {

    $plugin = $this->queueManager->createInstance('swapcard_content_queue_swapcard_event');
    $field_ids = [];
    $update_fields = [];
    $existing_fields = [];

    foreach ($node_types as $bundle => $fields) {

      $update_fields[$bundle] = $fields;

      if (!empty($default_fields)) {
        foreach ($default_fields as $field_name => $field_value) {
          if (isset($fields[$field_name])) {
            if (!empty($field_value)) {
              // Here we specifically update main config/settings with injecting
              // "id" value to each of the standard fields of a kind.
              $update_fields[$bundle][$field_name]['id'] = $field_value;

              $field_id = 'node.' . $bundle . '.' . $field_name;
              $field_config = $this->entityTypeManager->getStorage('field_config')->load($field_id);
              // $field_value;
              $field_ids[$field_id] = $field_config->getFieldStorageDefinition();
            }
          }
        }
        // Take care of custom fields now.
        foreach ($fields as $drupal_field_name => $swapcard_field) {
          if (isset($swapcard_field['custom']) && isset($swapcard_field['id'])) {
            $field_id = 'node.' . $bundle . '.' . $drupal_field_name;
            $existing_fields[$field_id] = $swapcard_field['id'];
          }
        }
      }

      // REVERSE check on previously exitsing (but not anymore) fields.
      $current_fields = [];
      if (!empty($filtered_fields)) {
        foreach ($filtered_fields as $filtered_field) {
          $current_fields[$filtered_field['drupal_field_name']] = $filtered_field['swapcard_field_id'];
        }
      }

      $diff = array_diff(array_keys($existing_fields), array_keys($current_fields));

      if (!empty($diff)) {
        foreach ($diff as $diff_field_id) {
          $field_name_array = strpos($diff_field_id, '.') !== FALSE ? explode('.', $diff_field_id) : [];
          $drupal_field_name = count($field_name_array) > 2 ? array_pop($field_name_array) : NULL;
          $field_bundle = count($field_name_array) > 1 ? array_pop($field_name_array) : NULL;
          if ($drupal_field_name && $field_bundle) {
            if (isset($fields[$drupal_field_name]['custom']) && isset($fields[$drupal_field_name]['id'])) {
              unset($update_fields[$field_bundle][$drupal_field_name]);
            }
          }
        }
      }

      if (!empty($filtered_fields)) {

        foreach ($filtered_fields as $filtered_field) {

          if (isset($filtered_field['drupal_field_name']) && !empty($filtered_field['drupal_field_name']) && isset($filtered_field['swapcard_field_id']) && !empty($filtered_field['swapcard_field_id'])) {

            $field_name_array = strpos($filtered_field['drupal_field_name'], '.') !== FALSE ? explode('.', $filtered_field['drupal_field_name']) : [];
            $drupal_field_name = count($field_name_array) > 2 ? array_pop($field_name_array) : NULL;
            $field_bundle = count($field_name_array) > 1 ? array_pop($field_name_array) : NULL;

            if ($drupal_field_name && $field_bundle) {

              $exists = in_array($drupal_field_name, array_keys($fields));
              $is_custom = isset($fields[$drupal_field_name]['custom']) && isset($fields[$drupal_field_name]['id']);

              if ($exists) {
                if (!$is_custom) {
                  $error = $this->t('Field @field_name cannot be defined as custom, it belongs to an array of default included fields.', [
                    '@field_name' => $drupal_field_name,
                  ]);
                  $plugin->messenger->addError($error);
                }
                else {
                  $this->processField($filtered_field, $update_fields, $field_ids);
                }
              }
              else {
                $this->processField($filtered_field, $update_fields, $field_ids);
              }
            }
          }
        }
      }
    }

    return [
      'update_fields' => $update_fields,
      'field_ids' => $field_ids,
    ];
  }

  /**
   * Add fieldd's id or a complete custom field to config.
   *
   * @todo this one.
   *
   * @param array $custom_field
   *   An array with custom field's properties.
   * @param array $update_fields
   *   Reference array containing update data for node_types config property.
   * @param array $field_ids
   *   Reference array of fields/definitions to refresh allowed_values.
   */
  private function processField(array $custom_field, array &$update_fields, array &$field_ids) {
    if ($field_config = $this->entityTypeManager->getStorage('field_config')->load($custom_field['drupal_field_name'])) {
      $update_fields[$field_config->getTargetBundle()][$field_config->getName()] = [
        'id' => $custom_field['swapcard_field_id'],
        'origin' => 'fields',
        'type' => $field_config->getType(),
        'custom' => TRUE,
      ];
      $texts = ['text', 'text_long', 'text_with_summary'];
      if (in_array($field_config->getType(), $texts)) {
        $update_fields[$field_config->getTargetBundle()][$field_config->getName()]['format_name'] = 'basic_html';
      }
      $field_ids[$custom_field['drupal_field_name']] = $field_config->getFieldStorageDefinition();
    }

  }

  /**
   * Add fieldd's id or a complete custom field to config.
   *
   * @todo this one.
   *
   * @param Drupal\Core\Form\FormStateInterface $form_state
   *   Form state object.
   * @param string $key
   *   A form element's key.
   * @param array $default_values
   *   An array with current vallues for the field (in config).
   *
   * @return int
   *   A current number of field items within the form.
   */
  private function multipleState(FormStateInterface $form_state, string $key, array $default_values = []) {

    // Gather the number of items in the form already.
    $num_items = $form_state->get('num_' . $key);
    // We have to ensure that there is at least one widget.
    if ($num_items === NULL) {
      if (!empty($default_values) && count($default_values) > 1) {
        $num_items = count($default_values);
      }
      else {
        $num_items = 1;
      }
    }

    $form_state->set('num_' . $key, $num_items);
    return $num_items;
  }

  /**
   * Append add/remove buttons to multiple fields.
   *
   * @todo this one.
   *
   * @param array $element
   *   Referenced form element, "swapcard_content_field" element in this case.
   * @param string $id
   *   Wrapper's unique id.
   * @param string $key
   *   Unique key of a field being manipulated.
   * @param string $label
   *   Label of a field being manipulated.
   * @param int $num_items
   *   Current number of field instances.
   */
  private function multipleOps(array &$element, string $id, string $key, string $label, int $num_items = 1) {

    $element['add_item'] = [
      '#type' => 'submit',
      '#value' => $this->t('Add @label', ['@label' => $label]),
      '#limit_validation_errors' => [],
      '#name' => 'op-' . $key,
      '#submit' => [
        [get_class($this), 'addItemSubmit'],
      ],
      '#weight' => 20,
      '#ajax' => [
        'callback' => [get_class($this), 'ajaxCallback'],
        'wrapper' => $id,
      ],
    ];

    if ($num_items > 1) {

      $element['remove_item'] = [
        '#type' => 'submit',
        '#value' => $this->t('Remove @label', ['@label' => $label]),
        '#limit_validation_errors' => [],
        '#name' => 'op-' . $key,
        '#submit' => [
          [get_class($this), 'removeItemSubmit'],
        ],
        '#weight' => 20,
        '#ajax' => [
          'callback' => [get_class($this), 'ajaxCallback'],
          'wrapper' => $id,
        ],
      ];
    }
  }

  /**
   * Process multiple Events form element.
   *
   * @todo this one.
   *
   * @param Drupal\Core\Form\FormStateInterface $form_state
   *   Form state object.
   * @param string $id
   *   Wrapper's unique id.
   * @param string $key
   *   Unique key of a field being manipulated.
   * @param string $label
   *   Label of a field being manipulated.
   *
   * @return array
   *   Processed form element.
   */
  private function multipleEvents(FormStateInterface $form_state, string $id, string $key, string $label) {

    $element = [];

    // Get current config/values.
    $config = $this->config('swapcard_content.settings');

    // Get curent form state's values.
    $values = $form_state->getValues();

    // Set and get current number of items.
    $num_items = $this->multipleState($form_state, $key, $config->get($key));

    // Container for multiple items.
    $element[$key] = [
      '#type' => 'container',
      '#tree' => TRUE,
      '#attributes' => [
        'id' => $id,
      ],
    ];

    for ($delta = 1; $delta <= $num_items; $delta++) {

      $index = $delta - 1;
      $item = NULL;

      if (is_array($config->get($key)) && !empty($config->get($key))) {
        $item = isset($config->get($key)[$index]) && !empty($config->get($key)[$index]) ? $config->get($key)[$index] : NULL;
      }
      else {
        if (isset($values[$key]) && !empty($values[$key])) {
          $item = isset($values[$key][$index]) && !empty($values[$key][$index]) ? $values[$key][$index] : NULL;
        }
      }

      // Child item.
      $item_id = Html::getId('swapcard-' . $key . '-' . $index);
      $element[$key][$index] = [
        '#type' => 'textfield',
        '#title' => $this->t('@label ID', ['@label' => $label]),
        '#default_value' => $item,
        '#attributes' => [
          'id' => $item_id,
        ],
      ];
    }

    // Render common add/remove ajax buttons.
    $this->multipleOps($element[$key], $id, $key, $label, $num_items);
    return $element;
  }

  /**
   * Process multiple Custom fields form element.
   *
   * @todo this one.
   *
   * @param Drupal\Core\Form\FormStateInterface $form_state
   *   Form state object.
   * @param string $id
   *   Wrapper's unique id.
   * @param string $key
   *   Unique key of a field being manipulated.
   * @param string $label
   *   Label of a field being manipulated.
   * @param array $fields
   *   Array of existing custom fields.
   *
   * @return array
   *   Processed form element.
   */
  private function multipleFields(FormStateInterface $form_state, string $id, string $key, string $label, array $fields = []) {

    $element = [];

    // Set and get current number of items.
    $num_items = $this->multipleState($form_state, $key, $fields);

    // Container for multiple items.
    $element[$key] = [
      '#type' => 'container',
      '#tree' => TRUE,
      '#attributes' => [
        'id' => $id,
      ],
    ];

    $item = [
      'drupal_field_name' => NULL,
      'swapcard_field_id' => NULL,
    ];

    for ($delta = 1; $delta <= $num_items; $delta++) {

      $index = $delta - 1;
      $item = [
        'drupal_field_name' => NULL,
        'swapcard_field_id' => NULL,
      ];

      if (!empty($fields) && isset($fields[$index]) && !empty($fields[$index])) {
        $item = $fields[$index];
      }

      if (isset($item['drupal_field_name'])) {
        $item['drupal_field_name'] = is_string($item['drupal_field_name']) ? $this->entityTypeManager->getStorage('field_config')->load($item['drupal_field_name']) : $item['drupal_field_name'];
      }

      // Child item.
      $element[$key][$index] = [
        '#type' => 'swapcard_content_custom_field',
        '#title' => $this->t('Custom field'),
        '#description' => $this->t('Handle custom created fields in Swapcard app.'),
        '#attributes' => [
          'id' => $id,
        ],
        '#default_value' => $item,
      ];
    }

    // Render common add/remove ajax buttons.
    $this->multipleOps($element[$key], $id, $key, $label, $num_items);
    return $element;

  }

}
