<?php

namespace Drupal\swapcard_content\Plugin\QueueWorker;

use Symfony\Component\Yaml\Yaml;
use Drupal\field\Entity\FieldStorageConfig;

/**
 * Creates QueueWorker for Swapcard fields.
 *
 * @QueueWorker(
 *   id = "swapcard_content_queue_swapcard_fields",
 *   title = @Translation("Swapcard Fields"),
 *   cron = {"time" = 90}
 * )
 */
class SwapcardQueueWorkerFields extends SwapcardQueueWorkerBase {

  /**
   * {@inheritdoc}
   */
  public function processItem($data) {

    if (!is_array($data) || empty($data)) {
      return;
    }

    foreach ($data as $field_id => $field_storage) {
      $options[$field_id] = $this->processFieldData($field_storage);
    }
    return $options;
  }

  /**
   * Generate field's allowed_values array.
   *
   * @param Drupal\field\Entity\FieldStorageConfig $definition
   *   Field config definition.
   * @param array $events
   *   An array with events ids.
   * @param bool $batch
   *   A flag for to determine if this is batch from UI or Cron run.
   *
   * @return array
   *   Associative array with key|Label pairs for select list type of field.
   */
  public function processFieldData(FieldStorageConfig $definition, array $events = [], bool $batch = FALSE) {

    $parent_config = $this->configFactory->get('swapcard.settings');
    $guzzle_options = $parent_config->get('guzzle_options');

    if ($parent_config->get('api_key') && !empty($guzzle_options['base_uri'])) {

      $events = !empty($events) ? $events : $this->config->get('events');

      // Note that Swapcard field's options (key|value)
      // are the same across all events.
      if (!empty($events) && isset($events[0]) && !empty($events[0])) {

        $field_id = $definition->id();
        $field_name = $definition->getName();

        $options = [];

        // Check on defualt fields first.
        foreach ($this->config->get('node_types') as $fields) {
          if (in_array($field_name, array_keys($fields))) {

            $field = $fields[$field_name];
            $field['events'] = $events;

            foreach ($events as $event_id) {

              $cid = 'swapcard_content:' . $event_id . ':' . $field_id;

              // Returned cached data. Especially worth when we speak about
              // node_edit forms where this request would run every time.
              if ($cached = $this->cache->get($cid)) {
                if (!empty($cached->data)) {
                  return $cached->data;
                }
              }

              // Run fields definition request for the event.
              $options[$field_id] = $this->fetchFieldData($event_id, $field, 'event', $batch);

              // Set this to cache.
              if (!empty($options[$field_id])) {
                $cache_tags = [$cid];
                $this->cache->set($cid, $options, $this->cache::CACHE_PERMANENT, $cache_tags);
              }
            }
          }
        }
        // Return array now, ready for list string like options.
        return $options;
      }
      // Events are not defined on config form, warn about it.
      else {
        $this->apiStatus('warning', 'events');
      }
    }
    // API key is missing on config form, warn about it.
    else {
      $this->apiStatus('error', 'api');
    }

  }

  /**
   * Get fields definitions from API.
   *
   * @param string $event_id
   *   Id of the event being processed.
   * @param array $field
   *   An array of values related to a field in drupal we match we remote.
   * @param string $swapcard_callback
   *   Swapcard API callback function name.
   * @param bool $batch
   *   A flag for to determine if this is batch from UI or Cron run.
   *
   * @return array
   *   Associative array with key|Label pairs for select list type of field.
   */
  protected function fetchFieldData(string $event_id, array $field, string $swapcard_callback, bool $batch = FALSE) {

    $parent_config = $this->configFactory->get('swapcard.settings');
    $guzzle_options = $parent_config->get('guzzle_options');

    $options = [];

    // TEST - add test response array to the queue.
    // Comes from YML file, instad of calling a remote API.
    if ($this->config->get('test') || $this->testPath) {

      $test_file_path = static::DUMP_PATH . '/swapcard_fields_' . $event_id . '_' . $field['id'] . '.test.yml';
      if (file_exists($test_file_path)) {
        $response['data'][$swapcard_callback] = Yaml::parseFile($test_file_path);
      }
      else {
        $this->messenger->addError($this->t('Specified test file <em>@path</em> does not exist.', [
          '@path' => $test_file_path,
        ]));
        return [];
      }
    }
    else {

      $parent_config = $this->configFactory->get('swapcard.settings');
      $guzzle_options = $parent_config->get('guzzle_options');

      $swapcard_plugin_manager = $this->swapcardPluginManager->createInstance('swapcard_fields', ['all']);
      $args = [
        'id' => '\"' . $event_id . '\"',
      ];

      // VERY ugly, but for now we do not have possibility to put suuch params
      // in annotation of Swapcard plugin.
      $query_string = $swapcard_plugin_manager->queryString($swapcard_callback, $args);
      $custom_query_string = 'fieldDefinitions(ids: ' . '\"' . $field['id'] . '\")';

      $callback_options = [
        'body' => str_replace('fieldDefinitions', $custom_query_string, $query_string),
      ];

      $response = $swapcard_plugin_manager->post($guzzle_options['base_uri'], $callback_options);
    }

    if (is_array($response) && isset($response['data']) && count($response['data']) > 0 && isset($response['data'][$swapcard_callback])) {

      // DUMP TEST file, if that's set on config.
      if ($this->config->get('dump')) {
        $this->dumpResponse($response['data'][$swapcard_callback], 'swapcard_fields_' . $event_id . '_' . $field['id'] . '.test.yml');
      }

      foreach ($response['data'] as $response_data) {
        if (isset($response_data['fieldDefinitions']) && !empty($response_data['fieldDefinitions'])) {

          foreach ($response_data['fieldDefinitions'] as $field_definition) {

            if (isset($field_definition['id']) && $field_definition['id'] == $field['id']) {

              switch ($field['type']) {

                case 'list_string':

                  if (isset($field_definition['optionsValues']) && !empty($field_definition['optionsValues'])) {

                    foreach ($field_definition['optionsValues'] as $optionsValue) {
                      if (isset($optionsValue['value']) && !empty($optionsValue['value'])) {

                        $options[$optionsValue['value']] = isset($optionsValue['translations']) && !empty($optionsValue['translations']) && isset($optionsValue['translations'][0]['value']) ? $optionsValue['translations'][0]['value'] : $optionsValue['value'];
                      }
                    }
                  }
                  break;
              }
            }
          }
        }
      }
    }

    return $options;
  }

}
